#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef USE_MINGW
#include <unistd.h>
#include <sys/time.h>
#endif

#include "netlib.h"

#ifdef USE_MINGW
void bcopy(const void *src, void *dst, unsigned int len)
{
  memcpy(dst, src, len);
}

int bcmp(const void *b1, const void *b2, unsigned int len)
{
  return memcmp(b1, b2, len);
}

int select(int fd, void *rfd, void *wfd, void *efd, struct timeval *timeout)
{
  struct timeval t;
  for (t.tv_sec = timeout->tv_sec; t.tv_sec; t.tv_sec--)
    usleep(1000000);
  usleep(timeout->tv_usec);
  return 0;
}
#endif

#if defined(USE_MINGW) || defined(__linux__)
unsigned int strlcpy(char *dst, const char *src, unsigned int size)
{
  strncpy(dst, src, size);
  if (size > 0)
    dst[size - 1] = '\0';
  return strlen(dst);
}
#endif

union val32 {
  uint32_t val32;
  uint8_t val8[4];
};

union val16 {
  uint16_t val16;
  uint8_t val8[2];
};

uint16_t htons(uint16_t h)
{
  union val16 n;
  n.val8[0] = (h >> 8) & 0xff;
  n.val8[1] = (h     ) & 0xff;
  return n.val16;
}

uint16_t ntohs(uint16_t n)
{
  union val16 h;
  h.val16 = n;
  return (h.val8[0] << 8) | h.val8[1];
}

uint32_t htonl(uint32_t h)
{
  union val32 n;
  n.val8[0] = (h >> 24) & 0xff;
  n.val8[1] = (h >> 16) & 0xff;
  n.val8[2] = (h >>  8) & 0xff;
  n.val8[3] = (h      ) & 0xff;
  return n.val32;
}

uint32_t ntohl(uint32_t n)
{
  union val32 h;
  h.val32 = n;
  return (h.val8[0] << 24) | (h.val8[1] << 16) |
    (h.val8[2] << 8) | h.val8[3];
}
