#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif
#else
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

static void help()
{
  printf("tcp-send <hostname> <port>\n");
  exit(0);
}

static void error_exit(char *message)
{
  fprintf(stderr, "%s", message);
  exit(1);
}

#ifdef WIN32
int winsock_init()
{
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    error_exit("WSAStartup() failed.\n");
  return 0;
}
#endif

int main(int argc, char *argv[])
{
  char *hostname;
  int port, s, readsize, sendsize, recvsize, writesize;
  int total_readsize = 0, total_sendsize = 0;
  int total_recvsize = 0, total_writesize = 0;
  struct sockaddr_in addr;
  socklen_t addrlen;
  char buffer[1024*80];
#ifdef USE_GETHOSTBYNAME
  struct hostent *host;
#endif

  argc--; argv++;
  if (argc < 2)
    help();

  hostname = *argv;
  argc--; argv++;
  port = atoi(*argv);
  argc--; argv++;

#ifdef WIN32
  winsock_init();
  setmode(0, O_BINARY);
#endif

#ifdef USE_GETHOSTBYNAME
  /* ۥ̾饢ɥ쥹 */
  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("Cannot resolve hostname.\n");
#endif

  /* åȤΥץ */
  s = socket(PF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("Cannot open socket.\n");

  /* ɥ쥹 */
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
#ifdef USE_GETHOSTBYNAME
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);
#else
  addr.sin_addr.s_addr = inet_addr(hostname);
#endif
  addrlen = sizeof(addr);

  /* Ф³ */
  if (connect(s, (struct sockaddr *)&addr, addrlen) < 0)
    error_exit("Cannot connect.\n");

  while (1) {
    /* ɸϤǡɤ߹ */
    readsize = read(0, buffer, sizeof(buffer));
    if (readsize < 0)
      error_exit("Cannot read.\n");
    if (readsize == 0)
      break;

    fprintf(stderr, "Read size from stdin:\t%d\n", readsize);
    total_readsize += readsize;

    /* ǡ */
#ifndef WIN32
    sendsize = write(s, buffer, readsize);
#else
    sendsize = send(s, buffer, readsize, 0);
#endif
    if (sendsize < 0)
      error_exit("Cannot send.\n");

    fprintf(stderr, "Send size on TCP:\t%d\n", sendsize);
    total_sendsize += sendsize;

    /* ǡμ */
#ifndef WIN32
    recvsize = read(s, buffer, sizeof(buffer));
#else
    recvsize = recv(s, buffer, sizeof(buffer), 0);
#endif
    if (recvsize < 0)
      error_exit("Cannot recv.\n");
    if (recvsize == 0)
      break;

    fprintf(stderr, "Recv size on TCP:\t%d\n", recvsize);
    total_recvsize += recvsize;

    /* ɸϤ˥ǡ񤭽Ф */
    writesize = write(1, buffer, recvsize);
    if (writesize < 0)
      error_exit("Cannot write.\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "Write size to stdout:\t%d\n", writesize);
    total_writesize += writesize;
  }

  fprintf(stderr, "\n");
  fprintf(stderr, "Total read size:\t%d\n", total_readsize);
  fprintf(stderr, "Total send size:\t%d\n", total_sendsize);
  fprintf(stderr, "Total recv size:\t%d\n", total_recvsize);
  fprintf(stderr, "Total write size:\t%d\n", total_writesize);

  /* åȤΥ */
#ifndef WIN32
  close(s);
#else
  closesocket(s);
#endif

  return 0;
}
