#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif
#else
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#ifdef WIN32
#ifndef USE_GETHOSTBYNAME
#ifndef USE_NETLIB
#warning Cannot use inet_pton() on Windows?
#warning If an error as "undefined reference to inet_pton" occurred
#warning on the link stage, Please try to use netlib as follows.
#warning $ unzip netlib-1.0.zip ; mv netlib-1.0 netlib ; make
#else
int inet_pton(int af, const char *src, void *dst);
#endif
#endif
#endif

static void help()
{
  printf("udp6-send <hostname> <port>\n");
  exit(0);
}

static void error_exit(char *message)
{
  fprintf(stderr, "%s", message);
  exit(1);
}

#ifdef WIN32
int winsock_init()
{
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    error_exit("WSAStartup() failed.\n");
  return 0;
}
#endif

int main(int argc, char *argv[])
{
  char *hostname;
  int port, s, readsize, sendsize, offset;
  struct sockaddr_in6 addr;
  socklen_t addrlen;
  char buffer[1024*80];
#ifdef USE_GETHOSTBYNAME
  struct hostent *host;
#endif

  argc--; argv++;
  if (argc < 2)
    help();

  hostname = *argv;
  argc--; argv++;
  port = atoi(*argv);
  argc--; argv++;

#ifdef WIN32
  winsock_init();
  setmode(0, O_BINARY);
#endif

#ifdef USE_GETHOSTBYNAME
  /* ۥ̾饢ɥ쥹 */
  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("Cannot resolve hostname.\n");
#endif

  /* åȤΥץ */
  s = socket(PF_INET6, SOCK_DGRAM, 0);
  if (s < 0)
    error_exit("Cannot open socket.\n");

  /* ɥ쥹 */
  memset(&addr, 0, sizeof(addr));
  addr.sin6_family = AF_INET6;
  addr.sin6_port = htons(port);
#ifdef USE_GETHOSTBYNAME
  memcpy(&addr.sin6_addr, host->h_addr, host->h_length);
#else
  if (inet_pton(AF_INET6, hostname, &addr.sin6_addr) < 1)
    error_exit("Invalid address.\n");
#endif
  addrlen = sizeof(addr);

  /* ɸϤǡɤ߹ */
  offset = 0;
  while (1) {
    readsize = read(0, buffer + offset, sizeof(buffer) - offset);
    if (readsize < 0)
      error_exit("Cannot read.\n");
    if (readsize == 0)
      break;
    offset += readsize;
    if (offset == sizeof(buffer))
      error_exit("Out of buffer.\n");
  }
  readsize = offset;

  fprintf(stderr, "Read size from stdin:\t%d\n", readsize);

  /* ǡ */
  sendsize = sendto(s, buffer, readsize, 0, (struct sockaddr *)&addr, addrlen);
  if (sendsize < 0)
    error_exit("Cannot send.\n");

  fprintf(stderr, "Send size on UDP6:\t%d\n", sendsize);

  /* åȤΥ */
#ifndef WIN32
  close(s);
#else
  closesocket(s);
#endif

  return 0;
}
